"""
    The "templates" folder stores HTML templates for native biothings
    web handlers and is structured as a dummy module to facilitate location
    resolution. Typically, ``biothings.web.templates.__path__[0]`` returns
    a string representing the location of this folder on the file system.
    This folder is intended to be used internally by the SDK developer.
"""
