"""
    Generate a customized BioThings API given a supported database.

    biothings.web.launcher: Launch web applications in different environments.
    biothings.web.applications: HTTP web application over data services below.
    biothings.web.services & query: Data services built on top of connections.
    biothings.web.connections: Elasticsearch, MongoDB and SQL database access.

"""
